/*
 * A B O U T
 *
 * filter portable draughts notation
 * interactive mode to playback single games
 * (c) 2004 Peter Chiocchetti <pch(a)myzel.net>
 *
 * This is the command line interface.
 */

/*
 * C O P Y I N G
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include "pdntool.h"

// lex/yacc
extern int yyparse (PDNgame *);
extern int yydebug;

// macros
#define destroy(ptr) if(ptr) free(ptr)
#define tocoor(x, y) (coor) {x, y}

// readview
#define PREV  -1
#define NEXT  -2
#define START -3
#define PART  -4
#define QUIT  -5
#define VIEW  -6

// printpdn
#define LLEN 74			// if more, break line

// globals
int verify = 1;			// validate moves
int pretty = 1;			// pretty print all of input
int matrix = 0;			// matrix mode
int errors = 0;			// pretty print failed games
int intact = 0;			// interactive mode
int guifmt = 0;			// -i xcheckers format

// types



/* batch output */

// print out a game in pretty pdn
static void
printpdn (PDNgame * pg)
{
  int n, llen;
  char temp[256];

  // reconstruct fen string
  if (pg->fen)
    {
      sprintf (temp, "%s:B", pg->turn);
      for (n = 1; n < SQUARE * SQUARE / 2; n++)
	{
	  if (pg->fpos[n] & BLACK)
	    {
	      if (pg->fpos[n] & MAN)
		sprintf (temp, "%s%d,", temp, n);
	      if (pg->fpos[n] & KING)
		sprintf (temp, "%sK%d,", temp, n);
	    }
	}
      temp[strlen (temp) - 1] = '\0';
      sprintf (temp, "%s:W", temp);
      for (n = 1; n < SQUARE * SQUARE / 2; n++)
	{
	  if (pg->fpos[n] & WHITE)
	    {
	      if (pg->fpos[n] & MAN)
		sprintf (temp, "%s%d,", temp, n);
	      if (pg->fpos[n] & KING)
		sprintf (temp, "%sK%d,", temp, n);
	    }
	}
      temp[strlen (temp) - 1] = '\0';
    }

  // print headers
  if (pg->event)
    printf ("[Event      \"%s\"]\n", pg->event);
  if (pg->date)
    printf ("[Date       \"%s\"]\n", pg->date);
  if (pg->black)
    printf ("[Black      \"%s\"]\n", pg->black);
  if (pg->white)
    printf ("[White      \"%s\"]\n", pg->white);
  if (pg->site)
    printf ("[Site       \"%s\"]\n", pg->site);
  if (pg->result)
    printf ("[Result     \"%s\"]\n", pg->result);
  if (pg->round)
    printf ("[Round      \"%s\"]\n", pg->round);
  if (pg->gametype)
    printf ("[GameType   \"%d\"]\n", pg->gametype);
  if (pg->fen)
    printf ("[FEN        \"%s\"]\n", temp);
  if (pg->ma[0])
    printf ("{%s}\n", pg->ma[0]);

  // print movelist, keep lines below LLEN chars
  for (n = 1, llen = 0; n < pg->nm; n++)
    {
      // move number
      if (n % 2 && llen + snprintf (temp, 6, "%d. ", n / 2 + 1) > LLEN)
	{
	  putchar ('\n');
	  llen = 0;
	}
      if (n % 2)
	{
	  llen += printf ("%d. ", n / 2 + 1);
	}
      // move, strength indicator
      if (llen + strlen (pg->mv[n]) + 1 > LLEN)
	{
	  putchar ('\n');
	  llen = 0;
	}
      if (pg->ms[n])
	llen += printf ("%s%s ", pg->mv[n], pg->ms[n]);
      else
	llen += printf ("%s ", pg->mv[n]);
      // annotation
      if (pg->ma[n] && llen + strlen (pg->ma[n]) > LLEN)
	{
	  putchar ('\n');
	  llen = 0;
	}
      if (pg->ma[n])
	llen += printf ("{%s} ", pg->ma[n]);
    }
  // print result
  if (!pg->result)
    pg->result = strdup ("*");
  if (llen + strlen (pg->result) > LLEN)
    putchar ('\n');
  printf ("%s\n\n", pg->result);
}


// matrix mode
// a compact notation of all the board positions in a game
static void
printmatrix (PDNgame * pg)
{
  int x, y, z;			// row, column, number of them
  int n = 0;
  // figures: w=5, b=6, W=9, B=10
  // field val:       01234567890123456
  char const f[17] = "-    wb  WB      ";

  // gametype
  switch (pg->gametype)
    {
    case INTERNL:
      z = 10;
      break;
    case ENGLISH:
    case ITALIAN:
    case RUSSIAN:
    case MAFIERZ:
      z = 8;
      break;
    default:
      assert ("gametype in printmatrix");
      return;
    }

  // print boards
  for (n = 0; n < pg->nm; n++)
    {
      printf ("%3d | ", n);
      for (y = 0; y < z; y++)
	for (x = 0; x < z; x++)
	  printf ("%c", f[pg->b[n + 1][x][y]]);
      printf (" | %d-%d\n", pg->mh[n][1], pg->mh[n][pg->mh[n][0]]);
    }
}



/* Interactive Mode */

// read the viewpoint
static int
readview ()
{
  int done = 0;
  char buf[32];

  while (!done)
    {
      printf ("\nViewpoint [black|white|part|quit] ? ");
      fgets (buf, sizeof (buf), stdin);

      if (buf[0] == '\n' || strstr (buf, "black"))
	return BLACK;
      if (strstr (buf, "white"))
	return WHITE;
      if (strstr (buf, "part") || strstr (buf, "break"))
	return PART;
      if (strstr (buf, "quit") || strstr (buf, "exit"))
	return QUIT;
    }
  return done;
}

// read commands
static int
readcommand ()
{
  int done = 0;
  char buf[32];

  while (!done)
    {
      printf ("\nCommand [next|prev|start|view|part|quit] ? ");
      fgets (buf, sizeof (buf), stdin);

      if (buf[0] == '\n' || strstr (buf, "next"))
	return NEXT;
      if (strstr (buf, "prev") || strstr (buf, "back"))
	return PREV;
      if (strstr (buf, "start"))
	return START;
      if (strstr (buf, "view"))
	return VIEW;
      if (strstr (buf, "part") || strstr (buf, "break"))
	return PART;
      if (strstr (buf, "quit") || strstr (buf, "exit"))
	return QUIT;
    }
  return done;
}

/*
 * ICDS style<5> 8x8 board string, 114 chars, 10x10 ->150 chars
 *

  0         1         2         3         4         5         
  012345678901234567890123456789012345678901234567890123456789
  <5>001eWWBlack Player NameWhite Player Nameb b b b  b b b bb
  6         7         8         9        10        11
  0123456789012345678901234567890123456789012345678901234
   b b b  - - - -- - - w  w w - ww w w w  w w w w00000000

  0         1         2         3         4         5         
  012345678901234567890123456789012345678901234567890123456789
  <5>001nBWBlack Player NameWhite Player Name-----------------
  6         7         8         9        10        11         
  012345678901234567890123456789012345678901234567890123456789
  ------------------------------------------------------------
  12       13        14        15
  0123456789012345678901234567890
  -----------------------00000000
 */

// print an icds style 5 board
static void
printicds (PDNgame * pg, int color, int view, int n)
{
  int x, y, z;			// row, column, number of them
  // figures: w=5, b=6, W=9, B=10
  // field val:       01234567890123456
  const char f[17] = "-    wb  WB      ";

  printf ("<5>");		// board display type
  printf ("%3d", pg->ng % 1000);	// game number, only three places
  switch (pg->gametype)
    {
    case INTERNL:
      z = 10;
      printf ("n");
      break;
    case ENGLISH:
      z = 8;
      printf ("e");
      break;
    case ITALIAN:
      z = 8;
      printf ("i");
      break;
    case RUSSIAN:
      z = 8;
      printf ("r");
      break;
    case MAFIERZ:
      z = 8;
      printf ("m");
      break;
    default:
      assert ("gametype in printicds");
      return;
    }
  view == BLACK ? printf ("B") : printf ("W");
  color == BLACK ? printf ("B") : printf ("W");
  view == BLACK ? printf ("%17s", pg->black) : printf ("%17s", pg->white);
  view == WHITE ? printf ("%17s", pg->black) : printf ("%17s", pg->white);

  for (y = 0; y < z; y++)
    for (x = 0; x < z; x++)
      printf ("%c", f[pg->b[n][x][y]]);

  printf ("000");		// black time
  printf ("000");		// white time
  printf ("\n");
}

// return board number for coordinates
static int
coortonumber (PDNgame * pg, coor c)
{
  // board coordinates are [y][x]!
  // INTERNL - lines mirrored?
  const int in[10][10] = {
    {5, 0, 4, 0, 3, 0, 2, 0, 1, 0},
    {0, 10, 0, 9, 0, 8, 0, 7, 0, 6},
    {15, 0, 14, 0, 13, 0, 12, 0, 11, 0},
    {0, 20, 0, 19, 0, 18, 0, 17, 0, 16},
    {25, 0, 24, 0, 23, 0, 22, 0, 21, 0},
    {0, 30, 0, 29, 0, 28, 0, 27, 0, 26},
    {35, 0, 34, 0, 33, 0, 32, 0, 31, 0},
    {0, 40, 0, 39, 0, 38, 0, 37, 0, 36},
    {45, 0, 44, 0, 43, 0, 42, 0, 41, 0},
    {0, 50, 0, 49, 0, 48, 0, 47, 0, 46},
  };

  // board coordinates are [y][x]!
  // ENGLISH
  const int en[8][8] = {
    {4, 0, 3, 0, 2, 0, 1, 0},
    {0, 8, 0, 7, 0, 6, 0, 5},
    {12, 0, 11, 0, 10, 0, 9, 0},
    {0, 16, 0, 15, 0, 14, 0, 13},
    {20, 0, 19, 0, 18, 0, 17, 0},
    {0, 24, 0, 23, 0, 22, 0, 21},
    {28, 0, 27, 0, 26, 0, 25, 0},
    {0, 32, 0, 31, 0, 30, 0, 29},
  };

  // ITALIAN
  const int it[8][8] = {
    {0, 29, 0, 30, 0, 31, 0, 32},
    {25, 0, 26, 0, 27, 0, 28, 0},
    {0, 21, 0, 22, 0, 23, 0, 24},
    {17, 0, 18, 0, 19, 0, 20, 0},
    {0, 13, 0, 14, 0, 15, 0, 16},
    {9, 0, 10, 0, 11, 0, 12, 0},
    {0, 5, 0, 6, 0, 7, 0, 8},
    {1, 0, 2, 0, 3, 0, 4, 0},
  };

  // RUSSIAN
  const int ru[8][8] = {
    {32, 0, 31, 0, 30, 0, 29, 0},
    {0, 28, 0, 27, 0, 26, 0, 25},
    {24, 0, 23, 0, 22, 0, 21, 0},
    {0, 20, 0, 19, 0, 18, 0, 17},
    {16, 0, 15, 0, 14, 0, 13, 0},
    {0, 12, 0, 11, 0, 10, 0, 9},
    {8, 0, 7, 0, 6, 0, 5, 0},
    {0, 4, 0, 3, 0, 2, 0, 1},
  };

  // MAFIERZ
  const int mf[8][8] = {
    {1, 0, 2, 0, 3, 0, 4, 0},
    {0, 5, 0, 6, 0, 7, 0, 8},
    {9, 0, 10, 0, 11, 0, 12, 0},
    {0, 13, 0, 14, 0, 15, 0, 16},
    {17, 0, 18, 0, 19, 0, 20, 0},
    {0, 21, 0, 22, 0, 23, 0, 24},
    {25, 0, 26, 0, 27, 0, 28, 0},
    {0, 29, 0, 30, 0, 31, 0, 32},
  };

  switch (pg->gametype)
    {
    case INTERNL:
      return in[c.y][c.x];
    case ENGLISH:
      return en[c.y][c.x];
    case ITALIAN:
      return it[c.y][c.x];
    case RUSSIAN:
      return ru[c.y][c.x];
    case MAFIERZ:
      return mf[c.y][c.x];
    default:
      printf ("Cannot get number from coordinate, unknown gametype %d.",
	      pg->gametype);
      return -1;
    }
}

// print an ascii board
static void
printascii (PDNgame * pg, int color, int view, int n)
{
  int x, y, z, r;		// row, column, rotation, number
  // figures: none=0, w=5, b=6, W=9, B=10
  // field val:       01234567890123456
  const char f[17] = "-    wb  WB      ";

  // rotate board, so the player sees it from her point of view
  r = 0;
  z = 8;
  if (pg->gametype == ENGLISH && view == BLACK)
    r = 1;
  if (pg->gametype == MAFIERZ && view == BLACK)
    r = 1;
  if (pg->gametype == ITALIAN && view == WHITE)
    r = 1;
  if (pg->gametype == RUSSIAN && view == WHITE)
    r = 1;

  for (y = 0; y < z; y++)
    {
      // board setup
      for (x = 0; x < z; x++)
	if (r)
	  printf (" %c", f[pg->b[n][x][z - y - 1]]);
	else
	  printf (" %c", f[pg->b[n][z - x - 1][y]]);
      printf ("\t");

      // board numbers
      for (x = 0; x < z; x++)
	if (r)
	  printf ("%2.0d", coortonumber (pg, tocoor (x, z - y - 1)));
	else
	  printf ("%2.0d", coortonumber (pg, tocoor (z - x - 1, y)));
      printf ("\n");
    }
}


// single step game
static void
playgame (PDNgame * pg)
{
  static int view = BLACK;	// view
  int cmd = NEXT;		// prev, next, ...
  int n = 0, m;
  int color = 0, startcolor;

  switch (pg->turn[0])
    {
    case 'b':
    case 'B':
      color = BLACK;
      break;
    case 'w':
    case 'W':
      color = WHITE;
      break;
    default:
      assert ("Color in playgame()");
    }
  startcolor = color;

  if (!pg->black)
    pg->black = strdup ("Black");
  if (!pg->white)
    pg->white = strdup ("White");

  view = readview ();
  if (view == QUIT)
    exit (0);
  if (view == PART)
    return;

  // play the game
  while (n < pg->nm)
    {
      m = n;			// latest move

      // board
      putchar ('\n');
      if (guifmt)
	printicds (pg, color, view, n + 1);
      printascii (pg, color, view, n + 1);
      putchar ('\n');

      // start annotation
      if (n == 0 && pg->ma[0])
	printf ("\n%s\n", pg->ma[0]);

      // move
      if (pg->ms[n])
	printf ("\n(%3d: %s%s)", n, pg->mv[n], pg->ms[n]);
      else
	printf ("\n(%3d: %s)", n, pg->mv[n]);
      if (pg->ma[n])
	printf (" %s ", pg->ma[n]);
      putchar ('\n');

      // result
      if (n + 1 == pg->nm)
	{
	  if (strstr (pg->result, "1-0"))
	    {
	      if (pg->gametype == ENGLISH)
		printf ("End of game, black wins.\n");
	      else
		printf ("End of game, white wins.\n");
	    }
	  if (strstr (pg->result, "1/2"))
	    printf ("End of game, its a draw.\n");
	  if (strstr (pg->result, "0-1"))
	    {
	      if (pg->gametype == ENGLISH)
		printf ("End of game, white wins.\n");
	      else
		printf ("End of game, black wins.\n");
	      if (strstr (pg->result, "*"))
		printf ("End of game, its not decided.\n");
	    }
	}

      // prompt
      cmd = readcommand ();
      if (cmd == PREV)
	if (n > 0 && m == n)
	  n--;
      if (cmd == NEXT)
	if (n + 1 < pg->nm)
	  n++;
      if (cmd == START)
	{
	  color = startcolor;
	  n = 0;
	  continue;
	}
      if (cmd == VIEW)
	view = view ^ CC;
      if (cmd == PART)
	n = pg->nm + 1;
      if (cmd == QUIT)
	exit (0);

      // change color
      if (m != n)
	color = color ^ CC;
    }
}



/* board setup */

/* shown from the side of the player to make the first move
   coordinates as in CB-api board paramenter

   int board[x][y]; struct coor (x, y); // x = col, y = row
 
   ENGLISH:

   row                           row

   7     w   w   w   w           7      32  31  30  29
   6   w   w   w   w             6    28  27  26  25
   5     w   w   w   w           5      24  23  22  21
   4   -   -   -   -             4    20  19  18  17
   3     -   -   -   -           3      16  15  14  13
   2   b   b   b   b             2    12  11  10   9
   1     b   b   b   b           1       8   7   6   5
   0   b   b   b   b             0     4   3   2   1

       0 1 2 3 4 5 6 7  col            0 1 2 3 4 5 6 7  col
 */

// initialize board to the starting position
static void
newboard_english (int b[SQUARE][SQUARE])
{
  int x, y, r = 0;		// row, column, rotation

  // empty board
  for (y = 0; y < 8; y++)
    for (x = 0; x < 8; x++)
      if ((x + y) % 2 == r)
	b[x][y] = EMPTY;
      else
	b[x][y] = FREE;

  // setup from start
  b[0][0] = b[2][0] = b[4][0] = b[6][0] =
    b[1][1] = b[3][1] = b[5][1] = b[7][1] =
    b[0][2] = b[2][2] = b[4][2] = b[6][2] = BLACK | MAN;
  b[1][7] = b[3][7] = b[5][7] = b[7][7] =
    b[0][6] = b[2][6] = b[4][6] = b[6][6] =
    b[1][5] = b[3][5] = b[5][5] = b[7][5] = WHITE | MAN;
}

// initialize board from fen
static void
fen2board_english (int b[SQUARE][SQUARE], int p[SQUARE * SQUARE / 2 + 1])
{
  int x, y, r = 0;		// counter, row, col, rotation

  // empty board
  for (y = 0; y < 8; y++)
    for (x = 0; x < 8; x++)
      if ((x + y) % 2 == r)
	b[x][y] = EMPTY;
      else
	b[x][y] = FREE;

  // setup from fen
  b[0][0] = p[4];
  b[2][0] = p[3];
  b[4][0] = p[2];
  b[6][0] = p[1];
  b[1][1] = p[8];
  b[3][1] = p[7];
  b[5][1] = p[6];
  b[7][1] = p[5];
  b[0][2] = p[12];
  b[2][2] = p[11];
  b[4][2] = p[10];
  b[6][2] = p[9];
  b[1][3] = p[16];
  b[3][3] = p[15];
  b[5][3] = p[14];
  b[7][3] = p[13];
  b[0][4] = p[20];
  b[2][4] = p[19];
  b[4][4] = p[18];
  b[6][4] = p[17];
  b[1][5] = p[24];
  b[3][5] = p[23];
  b[5][5] = p[22];
  b[7][5] = p[21];
  b[0][6] = p[28];
  b[2][6] = p[27];
  b[4][6] = p[26];
  b[6][6] = p[25];
  b[1][7] = p[32];
  b[3][7] = p[31];
  b[5][7] = p[30];
  b[7][7] = p[29];
}



/* validation, output completion */

// replay a game from parsed pdn
// return 0 on failure
static int
checkgame (PDNgame * pg)
{
  int b[SQUARE][SQUARE];	// current board
  int color;			// current turn
  int value = 0;		// validation result
  int i, n = 0;
  char temp[LLEN];
  CBmove cbmove;		// CB api

  // zero out board
  memset (&b, 0, sizeof (b));

  // default to english rules
  if (!pg->gametype)
    pg->gametype = ENGLISH;

  switch (pg->gametype)
    {
    case ENGLISH:
      if (pg->fen)
	fen2board_english (b, pg->fpos);
      else
	{
	  newboard_english (b);
	  pg->turn = strdup ("B");
	}
      break;
    default:
      fprintf (stderr, "%s: unsupported gametype %d\n", progname,
	       pg->gametype);
      return 0;
    }

  switch (pg->turn[0])
    {
    case 'b':
    case 'B':
      color = BLACK;
      break;
    case 'w':
    case 'W':
      color = WHITE;
      break;
    default:
      fprintf (stderr, "%s: unsupported turn %s\n", progname, pg->turn);
      return 0;
    }

  // check the game
  for (n = 1; n < pg->nm; n++)
    {
      // save board, before moving
      memcpy (pg->b[n], b, sizeof (b));

      // split move
      pg->mh[n][0] =
	sscanf (pg->mv[n], "%d%*c%d%*c%d%*c%d%*c%d%*c%d%*c%d%*c%d",
		&pg->mh[n][1], &pg->mh[n][2], &pg->mh[n][3], &pg->mh[n][4],
		&pg->mh[n][5], &pg->mh[n][6], &pg->mh[n][7], &pg->mh[n][8]);

      // validate move
      switch (pg->gametype)
	{
	case ENGLISH:
	  // check move, invalid = 0
	  value = islegal_english (b, color, pg->mh[n], &cbmove);
	default:
	  assert ("unsupported gametype\n");
	}

      if (!value)
	{
	  // annotate
	  snprintf (temp, LLEN, "next move %s found invalid by %s",
		    pg->mv[n], progname);
	  if (n >= 1 && pg->ma[n - 1])
	    {
	      // only if not already there
	      if (!strstr (pg->ma[n - 1], "invalid"))
		{
		  realloc (pg->ma[n - 1],
			   strlen (pg->ma[n - 1]) + strlen (temp) + 5);
		  strcat (pg->ma[n - 1], " - ");
		  strcat (pg->ma[n - 1], temp);
		}
	    }
	  else if (n >= 1)
	    pg->ma[n - 1] = strdup (temp);
	  else if (pg->ma[0])
	    {
	      // only if not already there
	      if (!strstr (pg->ma[0], "invalid"))
		{
		  realloc (pg->ma[0], strlen (pg->ma[0]) + strlen (temp) + 5);
		  strcat (pg->ma[0], " - ");
		  strcat (pg->ma[0], temp);
		}
	    }
	  else
	    pg->ma[0] = strdup (temp);

	  // print error
	  fprintf (stderr, "%s: invalid move found: %s (%d: %d-%d)\n",
		   progname, pg->mv[n], n / 2 + 1, pg->mh[n][1],
		   pg->mh[n][pg->mh[n][0]]);
	  fprintf (stderr, "%s: in game %d (before line %d, column %d).\n",
		   progname, pg->ng, line, col);
	  return value;
	}

      // adjust board
      b[cbmove.from.x][cbmove.from.y] = EMPTY;
      b[cbmove.to.x][cbmove.to.y] = cbmove.newpiece;
      for (i = 0; i < cbmove.jumps; i++)
	b[cbmove.del[i].x][cbmove.del[i].y] = EMPTY;

      // change color
      color = color ^ CC;
    }
  // save last board
  memcpy (pg->b[n], b, sizeof (b));
  return value;
}

// free memory allocated while parsing game
static void
pgfree (PDNgame * pg)
{
  int i;

  destroy (pg->event);
  destroy (pg->date);
  destroy (pg->black);
  destroy (pg->white);
  destroy (pg->site);
  destroy (pg->result);
  destroy (pg->round);
  destroy (pg->turn);
  for (i = 0; i <= MAXMOVES; i++)
    {
      destroy (pg->mv[i]);
      destroy (pg->ms[i]);
      destroy (pg->ma[i]);
    }
}



/* parser callback */

// a game is ready
void
gotgame (PDNgame * pg)
{
  int valid = 0;

  if (verify)
    valid = checkgame (pg);

  if (pretty)
    printpdn (pg);

  if (!valid && errors)
    printpdn (pg);

  if (matrix)
    printmatrix (pg);

  if (intact)
    playgame (pg);

  // clean up
  pgfree (pg);
  memset (pg, 0, sizeof (PDNgame));
  pg->mv[0] = strdup ("0-0");
  pg->nm = 1;
}



/* program setup */

int
main (int argc, char *argv[])
{
  char c;
  char *fn = 0;

  PDNgame pg;
  extern FILE *yyin;
  extern int errno;
  extern char *optarg;
  extern int optind, opterr, optopt;

  // program name
  if ((progname = strrchr (argv[0], '/')) == NULL)
    progname = argv[0];
  else
    progname++;

  // init pdngame info
  memset (&pg, 0, sizeof (PDNgame));
  pg.mv[0] = strdup ("0-0");
  pg.nm = 1;

  // commandline
  fprintf (stderr, "\n%s, filter portable draughts notation\n", progname);
  fprintf (stderr, "(c) 2004 Peter Chiocchetti <pch(a)myzel.net>\n");
  fprintf (stderr, "version: 2004-12-14 - zlib license\n");

  opterr = 0;
  while ((c = getopt (argc, argv, "eimnpvxh")) != EOF)
    {
      switch (c)
	{
	case 'e':
	  pretty = 0;
	  errors = 1;
	  break;
	case 'i':
	  intact = 1;
	  break;
	case 'm':
	  pretty = 0;
	  matrix = 1;
	  break;
	case 'n':
	  pretty = 0;
	  verify = 0;
	  break;
	case 'p':
	  pretty = 1;
	  break;
	case 'x':
	  intact = 1;
	  guifmt = 1;
	  break;
	case 'v':
	  yydebug = 1;
	  break;
	case 'h':
	case '?':
	  fprintf (stderr, "\nvalid options are:\n");
	  fprintf (stderr, "-p pretty print all games\n");
	  fprintf (stderr, "-e print only failed games\n");
	  fprintf (stderr, "-m select matrix output\n");
	  fprintf (stderr, "-n dont validate moves\n");
	  fprintf (stderr, "-i select interactive mode\n");
	  fprintf (stderr, "-x select xcheckers mode\n");
	default:
	  fprintf (stderr, "\nUnknown option `-%c'.\n", optopt);
	  exit (1);
	}
    }

  if (argc > optind)
    fn = argv[optind];

  // make stdout unbuffered
  // xcheckers needs that
  if (intact)
    setvbuf (stdout, NULL, _IONBF, 0);

  if (intact && !fn)
    {
      fprintf (stderr, "%s: in interactive mode, input must be a file!\n",
	       progname);
      return 0;
    }

  // parse file
  if (fn)
    {
      yyin = fopen (fn, "r");
      if (!yyin)
	{
	  fprintf (stderr, "%s: could not open file: %s (%s)\n",
		   progname, argv[2], strerror (errno));
	  exit (1);
	}
      yyparse (&pg);
      if (fclose (yyin))
	perror (progname);
    }
  else				// parse stdin
    yyparse (&pg);
  exit (0);
}
